package testutil

import (
	"git.cw.tr/mukan-network/mukan-sdk/testutil/configurator"
	_ "git.cw.tr/mukan-network/mukan-sdk/x/auth"           // import as blank for app wiring
	_ "git.cw.tr/mukan-network/mukan-sdk/x/auth/tx/config" // import as blank for app wiring
	_ "git.cw.tr/mukan-network/mukan-sdk/x/bank"           // import as blank for app wiring
	_ "git.cw.tr/mukan-network/mukan-sdk/x/consensus"      // import as blank for app wiring
	_ "git.cw.tr/mukan-network/mukan-sdk/x/distribution"   // import as blank for app wiring
	_ "git.cw.tr/mukan-network/mukan-sdk/x/genutil"        // import as blank for app wiring
	_ "git.cw.tr/mukan-network/mukan-sdk/x/mint"           // import as blank for app wiring
	_ "git.cw.tr/mukan-network/mukan-sdk/x/params"         // import as blank for app wiring
	_ "git.cw.tr/mukan-network/mukan-sdk/x/staking"        // import as blank for app wiring
)

var AppConfig = configurator.NewAppConfig(
	configurator.AuthModule(),
	configurator.BankModule(),
	configurator.StakingModule(),
	configurator.TxModule(),
	configurator.ConsensusModule(),
	configurator.ParamsModule(),
	configurator.GenutilModule(),
	configurator.DistributionModule(),
	configurator.MintModule(),
)
